/* The clock.cpp code is Copyright by Tom Horsley, 1998
 *
 * License is granted to anyone to freely distribute the code as long as the
 * copyright remains intact and I get paid too if you figure out some way to
 * make a profit off of this stuff :-).
 */

#include <windows.h>
#include <string.h>
#include "clock.h"

static char szAppName[] = "clock";

static char new_digits[7];
static char prev_digits[7];

/* geometry of clock face
 *
 * digits - x coordinates of the center of the digits to draw
 *
 * dots - x coordinates of the center of the dots to draw
 *
 * face_width - total width of the clock face
 *
 * face_digit - width of digit on clock face
 *
 * face_dot - width of dots on clock face
 */
static double digits[6] = {
   1.0,
   3.0,
   6.0,
   8.0,
   11.0,
   13.0
};
static double dots[2] = {
   4.5,
   9.5
};
static double face_width = 14.0;
static double face_digit = 2.0;
static double face_dot = 1.0;

/* geometry of segments in a single digit.
 *
 * segments - array giving center coordinates of each digit segment relative
 * to center of digit itself.
 *
 * digit_height, digit_width - total height and width of digit.
 *
 * segment_short, segment_long - short and long dimensions of rectangle
 * which holds each digit segment.
 */
static double segments[8][2] = {
   /* 1 */ { 0.0,  3.0},
   /* 2 */ {-1.5,  1.5},
   /* 3 */ { 1.5,  1.5},
   /* 4 */ { 0.0,  0.0},
   /* 5 */ {-1.5, -1.5},
   /* 6 */ { 1.5, -1.5},
   /* 7 */ { 0.0, -3.0}
};
static double digit_height = 7.0;
static double digit_width = 4.0;
static double segment_short = 1.0;
static double segment_long = 3.0;

/* geometry of the points that outline a single horizontal digit segment
 *
 * horizontal - array giving coordinates (relative to center) of a horizontal
 * digit segment.
 *
 * horizontal_width, horizontal_height - width and height of horizontal
 * segment.
 */
static double horizontal[6][2] = {
   {-2.0,  0.0},
   {-1.5,  0.5},
   { 1.5,  0.5},
   { 2.0,  0.0},
   { 1.5, -0.5},
   {-1.5, -0.5}
};
static double horizontal_width = 4.0;
static double horizontal_height = 1.3;

/* geometry of the points that outline a single vertical digit segment
 *
 * vertical - array giving coordinates (relative to center) of a vertical
 * digit segment.
 *
 * vertical_width, vertical_height - width and height of vertical
 * segment.
 */
static double vertical[6][2] = {
   { 0.0, -2.0},
   {-0.5, -1.5},
   {-0.5,  1.5},
   { 0.0,  2.0},
   { 0.5,  1.5},
   { 0.5, -1.5}
};
static double vertical_width = 1.3;
static double vertical_height = 4.0;

/* Index segment_map with digit we want to draw, get bit mask with 1 bit set
 * where segment should be lit, 0 where should be dark. Segment 1 is bit
 * (1 << 1), segment 2 is bit (1 << 2), etc.
 */
static unsigned char segment_map [10] = {
   (1 << 1)|(1 << 2)|(1 << 3)|(1 << 5)|(1 << 6)|(1 << 7),               /* 0 */
   (1 << 3)|(1 << 6),                                                   /* 1 */
   (1 << 1)|(1 << 3)|(1 << 4)|(1 << 5)|(1 << 7),                        /* 2 */
   (1 << 1)|(1 << 3)|(1 << 4)|(1 << 6)|(1 << 7),                        /* 3 */
   (1 << 2)|(1 << 3)|(1 << 4)|(1 << 6),                                 /* 4 */
   (1 << 1)|(1 << 2)|(1 << 4)|(1 << 6)|(1 << 7),                        /* 5 */
   (1 << 1)|(1 << 2)|(1 << 4)|(1 << 5)|(1 << 6)|(1 << 7),               /* 6 */
   (1 << 1)|(1 << 3)|(1 << 6),                                          /* 7 */
   (1 << 1)|(1 << 2)|(1 << 3)|(1 << 4)|(1 << 5)|(1 << 6)|(1 << 7),      /* 8 */
   (1 << 1)|(1 << 2)|(1 << 3)|(1 << 4)|(1 << 6),                        /* 9 */
};

static HPEN my_black_pen;
static HBRUSH my_cyan_brush;
static HBRUSH my_black_brush;
static HMENU hmenu;
static HINSTANCE hinst;

static POINT mouse_point;
static POINT window_point;

LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam);
void DrawAllDigits(HWND hwnd, HDC hdc, RECT* rect);
void DrawChangedDigits(HWND hwnd, HDC hdc, RECT* rect);
void DrawOneDigit(HWND hwnd, HDC hdc, RECT* rect, char dig);
void DrawHorizontalSegment(HWND hwnd, HDC hdc, RECT* rect);
void DrawVerticalSegment(HWND hwnd, HDC hdc, RECT* rect);
void FormatDigits(char * buf, WORD val);
void DrawDots(HWND hwnd, HDC hdc, RECT* rect);
void StashWindowPos(HWND hwnd);

static RECT init_pos;
static BOOL pos_changed;
static int ticks;

int WINAPI
WinMain(
   HINSTANCE hInstance,
   HINSTANCE hPrevInstance,
   PSTR szCmdLine,
   int iCmdShow) {
   HWND        hwnd;
   MSG         msg;
   WNDCLASSEX  wndclass;
   HKEY        posinfo;

   init_pos.top = 20;
   init_pos.left = 30;
   init_pos.bottom = 120;
   init_pos.right = 330;
   pos_changed = FALSE;
   ticks = 0;
   if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\NTCLI\\clock\\1.0",
                    0, KEY_READ, &posinfo) == ERROR_SUCCESS) {
      DWORD keytype;
      DWORD keyval;
      DWORD len;
      len = sizeof(keyval);
      if ((RegQueryValueEx(
              posinfo, "top", 0, &keytype, (LPBYTE)&keyval, &len) ==
                 ERROR_SUCCESS) && (keytype == REG_DWORD)) {
         init_pos.top = keyval;
      }
      len = sizeof(keyval);
      if ((RegQueryValueEx(
              posinfo, "left", 0, &keytype, (LPBYTE)&keyval, &len) ==
                 ERROR_SUCCESS) && (keytype == REG_DWORD)) {
         init_pos.left = keyval;
      }
      len = sizeof(keyval);
      if ((RegQueryValueEx(
              posinfo, "bottom", 0, &keytype, (LPBYTE)&keyval, &len) ==
                 ERROR_SUCCESS) && (keytype == REG_DWORD)) {
         init_pos.bottom = keyval;
      }
      len = sizeof(keyval);
      if ((RegQueryValueEx(
              posinfo, "right", 0, &keytype, (LPBYTE)&keyval, &len) ==
                 ERROR_SUCCESS) && (keytype == REG_DWORD)) {
         init_pos.right = keyval;
      }
   }
   hinst = hInstance;
   memset((void *)&prev_digits[0], 'X', sizeof(prev_digits));
   wndclass.cbSize = sizeof(wndclass);
   wndclass.style = CS_HREDRAW | CS_VREDRAW;
   wndclass.lpfnWndProc = WndProc;
   wndclass.cbClsExtra = 0;
   wndclass.cbWndExtra = 0;
   wndclass.hInstance = hInstance;
   wndclass.hIcon = LoadIcon(hInstance, "clockicon");
   wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
   wndclass.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
   wndclass.lpszMenuName = NULL;
   wndclass.lpszClassName = szAppName;
   wndclass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   RegisterClassEx(&wndclass);

   my_black_pen = CreatePen(PS_SOLID, 2, RGB(0,0,0));
   my_cyan_brush = CreateSolidBrush(RGB(0,255,255));
   my_black_brush = CreateSolidBrush(RGB(0,0,0));

   hwnd = CreateWindowEx(WS_EX_TOOLWINDOW,      // no taskbar entry!
                       szAppName,               // window class name
                       "Digital Clock",         // window caption
                       WS_POPUP | WS_SIZEBOX | WS_SYSMENU, // window style
                       init_pos.left,           // initial x position
                       init_pos.top,            // initial y position
                       init_pos.right - init_pos.left, // initial x size
                       init_pos.bottom - init_pos.top, // initial y size
                       NULL,                    // parent window handle
                       NULL,                    // window menu handle
                       hInstance,               // program instance handle
                       NULL) ;                  // creation parameters

   SetTimer(hwnd, 1, 333, NULL);
   ShowWindow(hwnd, iCmdShow);
   UpdateWindow(hwnd);

   while (GetMessage(&msg, NULL, 0, 0)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
   }
   return msg.wParam;
}

LRESULT CALLBACK
WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
   HDC         hdc;
   PAINTSTRUCT ps;
   RECT        rect;
   HGDIOBJ     old_pen;
   HGDIOBJ     old_brush;
   POINT       point;

   switch(iMsg) {
   case WM_CREATE :
      hmenu = LoadMenu(hinst, szAppName);
      hmenu = GetSubMenu(hmenu, 0);
      return 0;

   case WM_RBUTTONDOWN :
      point.x = LOWORD(lParam);
      point.y = HIWORD(lParam);
      ClientToScreen(hwnd, &point);
      TrackPopupMenu(hmenu, 0, point.x, point.y, 0, hwnd, NULL);
      return 0;

   case WM_COMMAND :
      switch(LOWORD(wParam)) {
      case IDM_EXIT :
         SendMessage(hwnd, WM_CLOSE, 0, 0);
         return 0;
      }
      break;

   case WM_PAINT :
      InvalidateRect(hwnd, NULL, FALSE);
      hdc = BeginPaint(hwnd, &ps);
      old_pen = SelectObject(hdc, my_black_pen);
      old_brush = SelectObject(hdc, my_cyan_brush);
      GetClientRect(hwnd, &rect);
      DrawAllDigits(hwnd, hdc, &rect);
      SelectObject(hdc, old_pen);
      SelectObject(hdc, old_brush);
      EndPaint(hwnd, &ps);
      return 0;

   case WM_TIMER :
      GetClientRect(hwnd, &rect);
      hdc = GetDC(hwnd);
      old_pen = SelectObject(hdc, my_black_pen);
      old_brush = SelectObject(hdc, my_cyan_brush);
      DrawChangedDigits(hwnd, hdc, &rect);
      SelectObject(hdc, old_pen);
      SelectObject(hdc, old_brush);
      ReleaseDC(hwnd, hdc);
      if (pos_changed && (ticks++ > 30)) {
         pos_changed = FALSE;
         ticks = 0;
         StashWindowPos(hwnd);
      }
      return 0;

   case WM_LBUTTONDOWN : {
      RECT cur;
      mouse_point.x = LOWORD(lParam);
      mouse_point.y = HIWORD(lParam);
      ClientToScreen(hwnd, &mouse_point);
      GetWindowRect(hwnd, &cur);
      window_point.x = cur.left;
      window_point.y = cur.top;
      SetCapture(hwnd);
      } return 0;

   case WM_LBUTTONUP :
      ReleaseCapture();
      return 0;

   case WM_MOUSEMOVE :
      if (wParam & MK_LBUTTON) {
         POINTS short_points;
         POINT new_mouse;
         POINT new_window;
         RECT cur;
         short_points = MAKEPOINTS(lParam);
         new_mouse.x = short_points.x;
         new_mouse.y = short_points.y;
         ClientToScreen(hwnd, &new_mouse);
         GetWindowRect(hwnd, &cur);
         new_window.x = window_point.x + (new_mouse.x - mouse_point.x);
         new_window.y = window_point.y + (new_mouse.y - mouse_point.y);
         MoveWindow(hwnd, new_window.x, new_window.y, cur.right - cur.left,
                    cur.bottom - cur.top, TRUE);
         mouse_point = new_mouse;
         window_point = new_window;
      }
      return 0;

   case WM_MOVE :
   case WM_SIZE :
      pos_changed = TRUE;
      return 0;

   case WM_DESTROY :
      if (pos_changed) {
         StashWindowPos(hwnd);
         pos_changed = FALSE;
      }
      DeleteObject(my_black_pen);
      DeleteObject(my_cyan_brush);
      DeleteObject(my_black_brush);
      KillTimer(hwnd, 1);
      PostQuitMessage(0);
      return 0;
   }

   return DefWindowProc(hwnd, iMsg, wParam, lParam);
}

void
DrawAllDigits(HWND hwnd, HDC hdc, RECT* rect) {
   memset((void *)&prev_digits[0], '*', sizeof(prev_digits));
   DrawChangedDigits(hwnd, hdc, rect);
   DrawDots(hwnd, hdc, rect);
}

void
DrawChangedDigits(HWND hwnd, HDC hdc, RECT* rect) {
   SYSTEMTIME right_now;
   GetLocalTime(&right_now);
   FormatDigits(&new_digits[0], right_now.wHour);
   FormatDigits(&new_digits[2], right_now.wMinute);
   FormatDigits(&new_digits[4], right_now.wSecond);
   if (new_digits[0] == '0') {
      new_digits[0] = ' ';
   }
   long pixel_width = rect->right - rect->left;
   long digit_pixels =
      (long)((((double)pixel_width) / face_width) * face_digit);
   long dot_pixels =
      (long)((((double)pixel_width) / face_width) * face_dot);
   long real_width = (6 * digit_pixels) + (2 * dot_pixels);
   long extra_dots = (pixel_width - real_width) / 2;
   rect->left += extra_dots;
   rect->right -= extra_dots;
   RECT dig_rect;
   double scale = ((double)real_width) / face_width;
   for (int i = 0; i < 6; ++i) {
      if (new_digits[i] != prev_digits[i]) {
         prev_digits[i] = new_digits[i];
         dig_rect.top = rect->top;
         dig_rect.bottom = rect->bottom;
         dig_rect.left = (long)((digits[i] - (face_digit / 2.0)) * scale);
         dig_rect.right = dig_rect.left + digit_pixels;
         dig_rect.top += 4;
         dig_rect.bottom -= 4;
         dig_rect.left += 4;
         dig_rect.right -= 4;
         DrawOneDigit(hwnd, hdc, &dig_rect, new_digits[i]);
      }
   }
}

void
DrawOneDigit(HWND hwnd, HDC hdc, RECT* rect, char dig) {
   SelectObject(hdc, my_black_brush);
   Rectangle(hdc, rect->left, rect->top, rect->right, rect->bottom);
   double cx = ((double)(rect->left + rect->right)) / 2.0;
   double cy = ((double)(rect->top + rect->bottom)) / 2.0;
   double hscale = ((double)(rect->right - rect->left)) / digit_width;
   double vscale = ((double)(rect->bottom - rect->top)) / digit_height;
   if ((dig >= '0') && (dig <= '9')) {
      unsigned char sm = segment_map[dig - '0'];
      SelectObject(hdc, my_cyan_brush);
      for (int i = 1; i <=7; ++i) {
         if (((1 << i) & sm) != 0) {
            RECT seg_rect;
            double sx = segments[i-1][0];
            double sy = segments[i-1][1];
            double px = cx + (hscale * sx);
            double py = cy - (vscale * sy);
            switch(i) {
            case 1:
            case 4:
            case 7: {
               double lng = (segment_long * hscale) / 2.0;
               double sht = (segment_short * vscale) / 2.0;
               seg_rect.top = (long)(py - sht);
               seg_rect.bottom = (long)(py + sht);
               seg_rect.left = (long)(px - lng);
               seg_rect.right = (long)(px + lng);
               DrawHorizontalSegment(hwnd, hdc, &seg_rect);
               } break;
            default: {
               double lng = (segment_long * vscale) / 2.0;
               double sht = (segment_short * hscale) / 2.0;
               seg_rect.top = (long)(py - lng);
               seg_rect.bottom = (long)(py + lng);
               seg_rect.left = (long)(px - sht);
               seg_rect.right = (long)(px + sht);
               DrawVerticalSegment(hwnd, hdc, &seg_rect);
               } break;
            }
         }
      }
   }
}

void
DrawHorizontalSegment(HWND hwnd, HDC hdc, RECT* rect) {
   POINT pts[6];
   double hscale = ((double)(rect->right - rect->left)) / horizontal_width;
   double vscale = ((double)(rect->bottom - rect->top)) / horizontal_height;
   double cx = ((double)(rect->left + rect->right)) / 2.0;
   double cy = ((double)(rect->top + rect->bottom)) / 2.0;
   for (int i = 0; i < 6; ++i) {
      pts[i].x = (long)(cx + (horizontal[i][0] * hscale));
      pts[i].y = (long)(cy - (horizontal[i][1] * vscale));
   }
   Polygon(hdc, pts, 6);
}

void
DrawVerticalSegment(HWND hwnd, HDC hdc, RECT* rect) {
   POINT pts[6];
   double hscale = ((double)(rect->right - rect->left)) / vertical_width;
   double vscale = ((double)(rect->bottom - rect->top)) / vertical_height;
   double cx = ((double)(rect->left + rect->right)) / 2.0;
   double cy = ((double)(rect->top + rect->bottom)) / 2.0;
   for (int i = 0; i < 6; ++i) {
      pts[i].x = (long)(cx + (vertical[i][0] * hscale));
      pts[i].y = (long)(cy - (vertical[i][1] * vscale));
   }
   Polygon(hdc, pts, 6);
}

void
FormatDigits(char * buf, WORD val) {
   buf[0] = (val / 10) + '0';
   buf[1] = (val % 10) + '0';
}

static double dot_box_height = 3.0;
static double dot_box_width = 2.0;
static double dot_box[4][2] = {
   { 0.0,  0.5},
   { 0.5,  0.0},
   { 0.0, -0.5},
   {-0.5,  0.0}
};

void
DrawDots(HWND hwnd, HDC hdc, RECT* rect) {
   POINT pts[4];
   long pixel_width = rect->right - rect->left;
   double scale = ((double)pixel_width) / face_width;
   long dot_pixels = (long)(scale * face_dot);
   RECT dot_rect;
   long half_height = (rect->bottom - rect->top) / 2;
   double hscale = ((double)dot_pixels) / dot_box_width;
   double vscale = ((double)half_height) / dot_box_height;
   for (int i = 0; i < 2; ++i) {
      dot_rect.left = (long)((dots[i] - (face_dot / 2.0)) * scale);
      dot_rect.right = dot_rect.left + dot_pixels;
      for (int j = 0; j < 2; ++j) {
         dot_rect.top = rect->top + (j * half_height);
         dot_rect.bottom = dot_rect.top + half_height;
         double cx = ((double)(dot_rect.left + dot_rect.right)) / 2.0;
         double cy = ((double)(dot_rect.top + dot_rect.bottom)) / 2.0;
         for (int k = 0; k < 4; ++k) {
            pts[k].x = (long)(cx + (dot_box[k][0] * hscale));
            pts[k].y = (long)(cy - (dot_box[k][1] * vscale));
         }
         Polygon(hdc, pts, 4);
      }
   }
}

void
StashWindowPos(HWND hwnd) {
   GetWindowRect(hwnd, &init_pos);
   HKEY posinfo;
   DWORD disp;
   if (RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\NTCLI\\clock\\1.0",
                      0, "", REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL,
                      &posinfo, &disp) == ERROR_SUCCESS) {
      DWORD keyval;
      keyval = init_pos.top;
      RegSetValueEx(posinfo, "top", 0, REG_DWORD,
                    (CONST BYTE *)&keyval, sizeof(keyval));
      keyval = init_pos.bottom;
      RegSetValueEx(posinfo, "bottom", 0, REG_DWORD,
                    (CONST BYTE *)&keyval, sizeof(keyval));
      keyval = init_pos.left;
      RegSetValueEx(posinfo, "left", 0, REG_DWORD,
                    (CONST BYTE *)&keyval, sizeof(keyval));
      keyval = init_pos.right;
      RegSetValueEx(posinfo, "right", 0, REG_DWORD,
                    (CONST BYTE *)&keyval, sizeof(keyval));
   }
}
